# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import fip64
from neutron_lib.api import extensions
from neutron_lib import constants as lib_const
from neutron_lib.db import utils as lib_db_utils
from neutron_lib.plugins import directory

from neutron.extensions import floatingip_pools as fip_pools_ext
from neutron.objects import base as base_obj
from neutron.objects import network as net_obj
from neutron.objects import subnet as subnet_obj


class FloatingIPPoolsDbMixin(object):
    """Class to support floating IP pool."""

    _is_v6_supported = None

    @staticmethod
    def _make_floatingip_pool_dict(context, subnet, fields=None):
        res = {'subnet_id': subnet.id,
               'subnet_name': subnet.name,
               'tenant_id': context.tenant_id,
               'network_id': subnet.network_id,
               'cidr': str(subnet.cidr)}

        return lib_db_utils.resource_fields(res, fields)

    def get_floatingip_pools(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        """Return information for available floating IP pools"""
        pager = base_obj.Pager(sorts, limit, page_reverse, marker)
        net_ids = [n.network_id
                   for n in net_obj.ExternalNetwork.get_objects(context)]
        # NOTE(hongbin): Use elevated context to make sure we have enough
        # permission to retrieve subnets that are not in current tenant
        # but belongs to external networks shared with current tenant.
        admin_context = context.elevated()
        subnet_objs = subnet_obj.Subnet.get_objects(admin_context,
                                                    _pager=pager,
                                                    network_id=net_ids)
        return [self._make_floatingip_pool_dict(context, obj, fields)
                for obj in subnet_objs
                if (obj.ip_version == lib_const.IP_VERSION_4 or
                    self.is_v6_supported)]

    @property
    def is_v6_supported(self):
        supported = self._is_v6_supported
        if supported is None:
            supported = False
            for plugin in directory.get_plugins().values():
                if extensions.is_extension_supported(plugin, fip64.ALIAS):
                    supported = True
                    break
        self._is_v6_supported = supported

        return supported


class FloatingIPPoolsMixin(FloatingIPPoolsDbMixin,
                           fip_pools_ext.FloatingIPPoolPluginBase):
    pass



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
from neutron_lib import constants


"""limit vlan allocation id values

Revision ID: dfe425060830
Revises: d8bdf05313f4
Create Date: 2020-04-27 16:58:33.110194

"""

# revision identifiers, used by Alembic.
revision = 'dfe425060830'
down_revision = 'd8bdf05313f4'


# NOTE(ralonsoh): "CHECK CONSTRAINT" is a feature heterogeneously implemented
# in different database engines and versions. For example:
# - MySQL: since version 8.0.16 (April 2019)
#   https://dev.mysql.com/doc/relnotes/mysql/8.0/en/news-8-0-16.html
# - MariaDB: since version 10.2.1 (July 2016)
#   https://mariadb.com/kb/en/mariadb-1021-release-notes/
# - PostgreSQL: since version 9.4 (December 2014)
#   https://www.postgresql.org/docs/9.4/ddl-constraints.html
#
# If the DB engine does not support yet this feature, it will be ignored. The
# VLAN tag constraint is enforced in the Neutron API. This extra enforcement
# in the DB engine wants to mimic the limitation imposed in the API side and
# avoid any spurious modification.

def upgrade():
    # https://docs.sqlalchemy.org/en/13/core/constraints.html#check-constraint
    constraint = ('vlan_id>=%(min_vlan)s AND vlan_id<=%(max_vlan)s' %
                  {'min_vlan': constants.MIN_VLAN_TAG,
                   'max_vlan': constants.MAX_VLAN_TAG})
    op.create_check_constraint(
        constraint_name='check_ml2_vlan_allocations0vlan_id',
        table_name='ml2_vlan_allocations',
        condition=constraint)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class LoggingApiBase(object, metaclass=abc.ABCMeta):
    """Logging API methods"""

    @abc.abstractmethod
    def create_log(self, context, log_obj):
        """Create a log_obj invocation.

        This method can be implemented by the specific driver subclass
        to update the backend where necessary with a specific log object.

        :param context: current running context information
        :param log_obj: a log objects being created

        """

    @abc.abstractmethod
    def create_log_precommit(self, context, log_obj):
        """Create a log_obj precommit.

        This method can be implemented by the specific driver subclass
        to handle the precommit event of a log_object that is being created.

        :param context: current running context information
        :param log_obj: a log object being created
        """

    @abc.abstractmethod
    def update_log(self, context, log_obj):
        """Update a log_obj invocation.

        This method can be implemented by the specific driver subclass
        to update the backend where necessary with a specific log object.

        :param context: current running context information
        :param log_obj: a log object being updated

        """

    @abc.abstractmethod
    def update_log_precommit(self, context, log_obj):
        """Update a log_obj precommit.

        This method can be implemented by the specific driver subclass
        to handle update precommit event of a log_object that is being updated.

        :param context: current running context information
        :param log_obj: a log_object being updated.
        """

    @abc.abstractmethod
    def delete_log(self, context, log_obj):
        """Delete a log_obj invocation.

        This method can be implemented by the specific driver subclass
        to delete the backend where necessary with a specific log object.

        :param context: current running context information
        :param log_obj: a log_object being deleted

        """

    @abc.abstractmethod
    def delete_log_precommit(self, context, log_obj):
        """Delete a log_obj precommit.

        This method can be implemented by the specific driver subclass
        to handle delete precommit event of a log_object that is being deleted.

        :param context: current running context information
        :param log_obj: a log_object being deleted
        """



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base


DEPRECATION_REASON = (
    "The Service Providers API now supports system scope and default roles.")

rules = [
    policy.DocumentedRuleDefault(
        name='get_service_provider',
        # NOTE: it can't be SYSTEM_OR_PROJECT_READER constant from the base
        # module because that is using "project_id" in the check string and the
        # service_provider resource don't belongs to any project thus such
        # check string would fail enforcment.
        check_str='role:reader',
        description='Get service providers',
        operations=[
            {
                'method': 'GET',
                'path': '/service-providers',
            },
        ],
        scope_types=['system', 'project'],
        deprecated_rule=policy.DeprecatedRule(
            name='get_service_provider',
            check_str=base.RULE_ANY,
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
]


def list_rules():
    return rules



# Copyright 2021 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from keystoneauth1 import exceptions as ks_exc
from neutron_lib import constants as n_const
from neutron_lib.placement import constants as placement_constants
from neutron_lib.placement import utils as placement_utils
from neutron_lib.plugins import constants as plugins_constants
from neutron_lib.plugins import directory
from neutron_lib.utils import helpers
from oslo_log import log as logging
from ovsdbapp.backend.ovs_idl import event as row_event

from neutron.agent.common import placement_report
from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils as ovn_utils
from neutron.common import utils as common_utils


LOG = logging.getLogger(__name__)


def _parse_ovn_cms_options(chassis):
    cms_options = ovn_utils.get_ovn_cms_options(chassis)
    return {n_const.RP_BANDWIDTHS: _parse_bandwidths(cms_options),
            n_const.RP_INVENTORY_DEFAULTS: _parse_inventory_defaults(
                cms_options),
            ovn_const.RP_HYPERVISORS: _parse_hypervisors(cms_options)}


def _parse_bridge_mappings(chassis):
    other_config = ovn_utils.get_ovn_chassis_other_config(chassis)
    bridge_mappings = other_config.get('ovn-bridge-mappings', '')
    bridge_mappings = helpers.parse_mappings(bridge_mappings.split(','),
                                             unique_values=False)
    return {k: [v] for k, v in bridge_mappings.items()}


def _parse_placement_option(option_name, cms_options):
    for cms_option in (cms_option for cms_option in cms_options if
                       option_name in cms_option):
        try:
            return cms_option.split('=')[1]
        except IndexError:
            break


def _parse_bandwidths(cms_options):
    bw_values = _parse_placement_option(n_const.RP_BANDWIDTHS, cms_options)
    if not bw_values:
        return {}

    return placement_utils.parse_rp_bandwidths(bw_values.split(';'))


def _parse_inventory_defaults(cms_options):
    inv_defaults = _parse_placement_option(n_const.RP_INVENTORY_DEFAULTS,
                                           cms_options)
    if not inv_defaults:
        return {}

    inventory = {}
    for inv_default in inv_defaults.split(';'):
        for key in placement_constants.INVENTORY_OPTIONS:
            if key in inv_default:
                inventory[key] = inv_default.split(':')[1]
    return placement_utils.parse_rp_inventory_defaults(inventory)


def _parse_hypervisors(cms_options):
    hyperv = _parse_placement_option(ovn_const.RP_HYPERVISORS, cms_options)
    if not hyperv:
        return {}

    return helpers.parse_mappings(hyperv.split(';'), unique_values=False)


def _send_deferred_batch(state):
    if not state:
        return

    deferred_batch = state.deferred_sync()
    for deferred in deferred_batch:
        try:
            LOG.debug('Placement client: %s', str(deferred))
            deferred.execute()
        except Exception:
            LOG.exception('Placement client call failed: %s', str(deferred))


def dict_chassis_config(state):
    if state:
        return {n_const.RP_BANDWIDTHS: state._rp_bandwidths,
                n_const.RP_INVENTORY_DEFAULTS: state._rp_inventory_defaults,
                ovn_const.RP_HYPERVISORS: state._hypervisor_rps}


class ChassisBandwidthConfigEvent(row_event.RowEvent):
    """Chassis create update event to track the bandwidth config changes."""

    def __init__(self, placement_extension):
        self._placement_extension = placement_extension
        # NOTE(ralonsoh): BW resource provider information is stored in
        # "Chassis", not "Chassis_Private".
        table = 'Chassis'
        events = (self.ROW_CREATE, self.ROW_UPDATE)
        super().__init__(events, table, None)
        self.event_name = 'ChassisBandwidthConfigEvent'

    def run(self, event, row, old):
        name2uuid = self._placement_extension.name2uuid()
        state = self._placement_extension.build_placement_state(row, name2uuid)
        if not state:
            return

        _send_deferred_batch(state)
        ch_config = dict_chassis_config(state)
        LOG.debug('OVN chassis %(chassis)s Placement configuration modified: '
                  '%(config)s', {'chassis': row.name, 'config': ch_config})


@common_utils.SingletonDecorator
class OVNClientPlacementExtension(object):
    """OVN client Placement API extension"""

    def __init__(self, driver):
        LOG.info('Starting OVNClientPlacementExtension')
        super().__init__()
        self._config_event = None
        self._reset(driver)

    def _reset(self, driver):
        """Reset the interval members values
        This class is a singleton. Once initialized, any other new instance
        will return the same object reference with the same member values.
        This method is used to reset all of them as when the class is initially
        instantiated if needed.
        """
        self._driver = driver
        self._placement_plugin = None
        self._plugin = None
        self.uuid_ns = ovn_const.OVN_RP_UUID
        self.supported_vnic_types = ovn_const.OVN_SUPPORTED_VNIC_TYPES
        if not self.enabled:
            return

        if not self._config_event:
            self._config_event = ChassisBandwidthConfigEvent(self)
            try:
                self._driver._sb_idl.idl.notify_handler.watch_events(
                    [self._config_event])
            except AttributeError:
                # "sb_idl.idl.notify_handler" is not present in the
                # MaintenanceWorker.
                pass

    @property
    def placement_plugin(self):
        if self._placement_plugin is None:
            self._placement_plugin = directory.get_plugin(
                plugins_constants.PLACEMENT_REPORT)
        return self._placement_plugin

    @property
    def enabled(self):
        return bool(self.placement_plugin)

    @property
    def plugin(self):
        if self._plugin is None:
            self._plugin = self._driver._plugin
        return self._plugin

    @property
    def ovn_mech_driver(self):
        if self._ovn_mech_driver is None:
            self._ovn_mech_driver = (
                self.plugin.mechanism_manager.mech_drivers['ovn'].obj)
        return self._ovn_mech_driver

    def get_chassis_config(self):
        """Read all Chassis BW config and returns the Placement states"""
        chassis = {}
        name2uuid = self.name2uuid()
        for ch in self._driver._sb_idl.chassis_list().execute(
                check_error=True):
            state = self.build_placement_state(ch, name2uuid)
            if state:
                chassis[ch.name] = state

        return chassis

    def read_initial_chassis_config(self):
        """Read the Chassis BW configuration and update the Placement API

        This method is called once from the MaintenanceWorker when the Neutron
        server starts.
        """
        if not self.enabled:
            return

        chassis = self.get_chassis_config()
        for state in chassis.values():
            _send_deferred_batch(state)
        msg = ', '.join(['Chassis %s: %s' % (name, dict_chassis_config(state))
                         for (name, state) in chassis.items()]) or '(no info)'
        LOG.debug('OVN chassis Placement initial configuration: %s', msg)
        return chassis

    def name2uuid(self, name=None):
        try:
            rps = self.placement_plugin._placement_client.\
                list_resource_providers(name=name)['resource_providers']
        except (ks_exc.HttpError, ks_exc.ClientException):
            LOG.warning('Error connecting to Placement API.')
            return {}

        _name2uuid = {rp['name']: rp['uuid'] for rp in rps}
        LOG.info('Placement information about resource providers '
                 '(name:uuid):%s ', _name2uuid)
        return _name2uuid

    def build_placement_state(self, chassis, name2uuid):
        bridge_mappings = _parse_bridge_mappings(chassis)
        cms_options = _parse_ovn_cms_options(chassis)
        LOG.debug('Building placement options for chassis %s: %s',
                  chassis.name, cms_options)
        hypervisor_rps = {}
        for device, hyperv in cms_options[ovn_const.RP_HYPERVISORS].items():
            try:
                hypervisor_rps[device] = {'name': hyperv,
                                          'uuid': name2uuid[hyperv]}
            except (KeyError, AttributeError):
                continue

        bridges = set(itertools.chain(*bridge_mappings.values()))
        # Remove "cms_options[RP_BANDWIDTHS]" not present in "hypervisor_rps"
        # and "bridge_mappings". If we don't have a way to match the RP bridge
        # with a host ("hypervisor_rps") or a way to match the RP bridge with
        # an external network ("bridge_mappings"), this value is irrelevant.
        rp_bw = cms_options[n_const.RP_BANDWIDTHS]
        if rp_bw:
            cms_options[n_const.RP_BANDWIDTHS] = {
                device: bw for device, bw in rp_bw.items() if
                device in hypervisor_rps and device in bridges}

        # NOTE(ralonsoh): OVN only reports min BW RPs; packet processing RPs
        # will be added in a future implementation. If no RP_BANDWIDTHS values
        # are present (that means there is no BW information for any interface
        # in this host), no "PlacementState" is returned.
        return placement_report.PlacementState(
            rp_bandwidths=cms_options[n_const.RP_BANDWIDTHS],
            rp_inventory_defaults=cms_options[n_const.RP_INVENTORY_DEFAULTS],
            rp_pkt_processing={},
            rp_pkt_processing_inventory_defaults=None,
            driver_uuid_namespace=self.uuid_ns,
            agent_type=ovn_const.OVN_CONTROLLER_AGENT,
            hypervisor_rps=hypervisor_rps,
            device_mappings=bridge_mappings,
            supported_vnic_types=self.supported_vnic_types,
            client=self.placement_plugin._placement_client)



# Copyright (c) 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys
from unittest import mock

from neutron_lib.agent import topics
from neutron_lib import constants
from neutron_lib.utils import helpers
from oslo_config import cfg
from oslo_service import service

from neutron.agent.linux import ip_lib
from neutron.common import config as common_config
from neutron.plugins.ml2.drivers.agent import _agent_manager_base as amb
from neutron.plugins.ml2.drivers.macvtap.agent import macvtap_neutron_agent
from neutron.plugins.ml2.drivers.macvtap import macvtap_common
from neutron.tests import base


INTERFACE_MAPPINGS = {'physnet1': 'eth1'}
NETWORK_ID = 'net-id123'
NETWORK_SEGMENT_VLAN = amb.NetworkSegment('vlan', 'physnet1', 1)
NETWORK_SEGMENT_FLAT = amb.NetworkSegment('flat', 'physnet1', None)


class TestMacvtapRPCCallbacks(base.BaseTestCase):
    def setUp(self):
        super(TestMacvtapRPCCallbacks, self).setUp()

        agent = mock.Mock()
        agent.mgr = mock.Mock()
        agent.mgr.interface_mappings = INTERFACE_MAPPINGS
        self.rpc = macvtap_neutron_agent.MacvtapRPCCallBack(mock.Mock(), agent,
                                                            mock.Mock())

    def test_network_delete_vlan(self):
        self.rpc.network_map = {NETWORK_ID: NETWORK_SEGMENT_VLAN}
        with mock.patch.object(ip_lib.IpLinkCommand, 'delete') as mock_del,\
                mock.patch.object(macvtap_common, 'get_vlan_device_name',
                                  return_value='vlan1'),\
                mock.patch.object(ip_lib.IPDevice, 'exists',
                                  return_value=True):
            self.rpc.network_delete("anycontext", network_id=NETWORK_ID)
            self.assertTrue(mock_del.called)

    def test_network_delete_flat(self):
        self.rpc.network_map = {NETWORK_ID: NETWORK_SEGMENT_FLAT}
        with mock.patch.object(ip_lib.IpLinkCommand, 'delete') as mock_del:
            self.rpc.network_delete(
                "anycontext", network_id=NETWORK_SEGMENT_FLAT.segmentation_id)
            self.assertFalse(mock_del.called)

    def test_port_update(self):
        port = {'id': 'port-id123', 'mac_address': 'mac1'}
        self.rpc.port_update(context=None, port=port)
        self.assertEqual(set(['mac1']), self.rpc.updated_devices)


class TestMacvtapManager(base.BaseTestCase):
    def setUp(self):
        super(TestMacvtapManager, self).setUp()
        with mock.patch.object(ip_lib, 'device_exists', return_value=True):
            self.mgr = macvtap_neutron_agent.MacvtapManager(INTERFACE_MAPPINGS)

    def test_validate_interface_mappings_dev_exists(self):
        good_mapping = {'physnet1': 'eth1', 'physnet2': 'eth2'}
        self.mgr.interface_mappings = good_mapping
        with mock.patch.object(ip_lib, 'device_exists', return_value=True)\
                as mock_de:
            self.mgr.validate_interface_mappings()
            mock_de.assert_any_call('eth1')
            mock_de.assert_any_call('eth2')
            self.assertEqual(2, mock_de.call_count)

    def test_validate_interface_mappings_dev_not_exists(self):
        bad_mapping = {'physnet1': 'foo'}
        self.mgr.interface_mappings = bad_mapping
        with mock.patch.object(ip_lib, 'device_exists', return_value=False)\
                as mock_de, mock.patch.object(sys, 'exit') as mock_exit:
            self.mgr.validate_interface_mappings()
            mock_de.assert_called_with('foo')
            mock_exit.assert_called_once_with(1)

    def _test_ensure_port_admin_state(self, admin_state):
        dev = 'macvtap1'
        mac = 'mac1'

        self.mgr.mac_device_name_mappings = {mac: dev}
        with mock.patch.object(ip_lib, 'IPDevice') as mock_ip_dev:
            self.mgr.ensure_port_admin_state(mac, admin_state)
            self.assertEqual(admin_state, mock_ip_dev(dev).link.set_up.called)
            self.assertNotEqual(admin_state,
                                mock_ip_dev(dev).link.set_down.called)

    def test_ensure_port_admin_state_up(self):
        self._test_ensure_port_admin_state(True)

    def test_ensure_port_admin_state_down(self):
        self._test_ensure_port_admin_state(False)

    def test_get_all_devices(self):
        listing = ['foo', 'macvtap0', 'macvtap1', 'bar']
        # set some mac mappings to make sure they are cleaned up
        self.mgr.mac_device_name_mappings = {'foo': 'bar'}
        with mock.patch.object(os, 'listdir',
                               return_value=listing) as mock_ld,\
                mock.patch.object(ip_lib, 'get_device_mac') as mock_gdn:
            mock_gdn.side_effect = ['mac0', 'mac1']

            result = self.mgr.get_all_devices()
            mock_ld.assert_called_once_with(macvtap_neutron_agent.MACVTAP_FS)
            self.assertEqual(set(['mac0', 'mac1']), result)
            self.assertEqual({'mac0': 'macvtap0', 'mac1': 'macvtap1'},
                             self.mgr.mac_device_name_mappings)

    def test_get_agent_configurations(self):
        expected = {'interface_mappings': INTERFACE_MAPPINGS}
        self.assertEqual(expected, self.mgr.get_agent_configurations())

    def test_get_agent_id_ok(self):
        mock_devices = [ip_lib.IPDevice('macvtap1')]
        with mock.patch.object(ip_lib.IPWrapper, 'get_devices',
                               return_value=mock_devices),\
            mock.patch.object(ip_lib, 'get_device_mac',
                              return_value='foo:bar'):
            self.assertEqual('macvtapfoobar', self.mgr.get_agent_id())

    def test_get_agent_id_fail(self):
        mock_devices = []
        with mock.patch.object(ip_lib.IPWrapper, 'get_devices',
                               return_value=mock_devices),\
                mock.patch.object(sys, 'exit') as mock_exit:
            self.mgr.get_agent_id()
            mock_exit.assert_called_once_with(1)

    def test_get_agent_id_no_mac(self):
        mock_devices = [ip_lib.IPDevice('macvtap0'),
                        ip_lib.IPDevice('macvtap1')]
        with mock.patch.object(ip_lib.IPWrapper, 'get_devices',
                               return_value=mock_devices),\
            mock.patch.object(ip_lib, 'get_device_mac',
                              side_effect=[None, 'foo:bar:1']) as mock_gdm:
            self.assertEqual('macvtapfoobar1', self.mgr.get_agent_id())
            mock_gdm.assert_has_calls([mock.call('macvtap0'),
                                       mock.call('macvtap1')])

    def test_get_extension_driver_type(self):
        self.assertEqual('macvtap', self.mgr.get_extension_driver_type())

    def test_get_rpc_callbacks(self):
        context = mock.Mock()
        agent = mock.Mock()
        sg_agent = mock.Mock()
        obj = self.mgr.get_rpc_callbacks(context, agent, sg_agent)
        self.assertIsInstance(obj, macvtap_neutron_agent.MacvtapRPCCallBack)

    def test_get_rpc_consumers(self):
        consumers = [[topics.PORT, topics.UPDATE],
                     [topics.NETWORK, topics.DELETE],
                     [topics.SECURITY_GROUP, topics.UPDATE]]
        self.assertEqual(consumers, self.mgr.get_rpc_consumers())

    def test_plug_interface(self):
        self.mgr.mac_device_name_mappings['mac1'] = 'macvtap0'
        with mock.patch.object(ip_lib.IpLinkCommand, 'set_allmulticast_on')\
                as mock_sao:
            self.mgr.plug_interface('network_id', 'network_segment', 'mac1',
                                    'device_owner')
            self.assertTrue(mock_sao.called)


class TestMacvtapMain(base.BaseTestCase):
    def test_parse_interface_mappings_good(self):
        cfg.CONF.set_override('physical_interface_mappings', 'good_mapping',
                              'macvtap')
        with mock.patch.object(helpers, 'parse_mappings',
                               return_value=INTERFACE_MAPPINGS):
            mappings = macvtap_neutron_agent.parse_interface_mappings()
            self.assertEqual(INTERFACE_MAPPINGS, mappings)

    def test_parse_interface_mappings_bad(self):
        cfg.CONF.set_override('physical_interface_mappings', 'bad_mapping',
                              'macvtap')
        with mock.patch.object(helpers, 'parse_mappings',
                               side_effect=ValueError('bad mapping')),\
                mock.patch.object(sys, 'exit') as mock_exit:
            macvtap_neutron_agent.parse_interface_mappings()
            mock_exit.assert_called_with(1)

    def test_parse_interface_mappings_no_mapping(self):
        with mock.patch.object(sys, 'exit') as mock_exit:
            macvtap_neutron_agent.parse_interface_mappings()
            mock_exit.assert_called_with(1)

    def test_validate_firewall_driver_noop_long(self):
        cfg.CONF.set_override('firewall_driver',
                              'neutron.agent.firewall.NoopFirewallDriver',
                              'SECURITYGROUP')
        macvtap_neutron_agent.validate_firewall_driver()

    def test_validate_firewall_driver_noop(self):
        cfg.CONF.set_override('firewall_driver',
                              'noop',
                              'SECURITYGROUP')
        macvtap_neutron_agent.validate_firewall_driver()

    def test_validate_firewall_driver_other(self):
        cfg.CONF.set_override('firewall_driver',
                              'foo',
                              'SECURITYGROUP')
        with mock.patch.object(sys, 'exit')as mock_exit:
            macvtap_neutron_agent.validate_firewall_driver()
            mock_exit.assert_called_with(1)

    def test_main(self):
        cfg.CONF.set_override('quitting_rpc_timeout', 1, 'AGENT')
        cfg.CONF.set_override('polling_interval', 2, 'AGENT')

        mock_manager_return = mock.Mock(spec=amb.CommonAgentManagerBase)
        mock_launch_return = mock.Mock()

        with mock.patch.object(common_config, 'init'),\
            mock.patch.object(common_config, 'setup_logging'),\
            mock.patch.object(service, 'launch',
                              return_value=mock_launch_return) as mock_launch,\
            mock.patch.object(macvtap_neutron_agent,
                              'parse_interface_mappings',
                              return_value=INTERFACE_MAPPINGS) as mock_pim,\
            mock.patch.object(macvtap_neutron_agent,
                              'validate_firewall_driver') as mock_vfd,\
            mock.patch('neutron.plugins.ml2.drivers.agent._common_agent.'
                       'CommonAgentLoop') as mock_loop,\
            mock.patch('neutron.plugins.ml2.drivers.macvtap.agent.'
                       'macvtap_neutron_agent.MacvtapManager',
                       return_value=mock_manager_return) as mock_manager:
            macvtap_neutron_agent.main()
            self.assertTrue(mock_vfd.called)
            self.assertTrue(mock_pim.called)
            mock_manager.assert_called_with(INTERFACE_MAPPINGS)
            mock_loop.assert_called_with(mock_manager_return, 2, 1,
                                         constants.AGENT_TYPE_MACVTAP,
                                         constants.AGENT_PROCESS_MACVTAP)
            self.assertTrue(mock_launch.called)
            self.assertTrue(mock_launch_return.wait.called)



# Copyright 2015 Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import port_security as psec
from neutron_lib.api import validators
from neutron_lib.plugins.ml2 import api
from neutron_lib.utils import net
from oslo_log import log as logging

from neutron.db import portsecurity_db_common as ps_db_common

LOG = logging.getLogger(__name__)


class PortSecurityExtensionDriver(api.ExtensionDriver,
                                  ps_db_common.PortSecurityDbCommon):
    _supported_extension_alias = 'port-security'

    def initialize(self):
        LOG.info("PortSecurityExtensionDriver initialization complete")

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def process_create_network(self, context, data, result):
        # Create the network extension attributes.
        if psec.PORTSECURITY not in data:
            data[psec.PORTSECURITY] = psec.DEFAULT_PORT_SECURITY
        self._process_network_port_security_create(context, data, result)

    def process_update_network(self, context, data, result):
        # Update the network extension attributes.
        if psec.PORTSECURITY in data:
            self._process_network_port_security_update(context, data, result)

    def process_create_port(self, context, data, result):
        # Create the port extension attributes.
        data[psec.PORTSECURITY] = self._determine_port_security(context, data)
        self._process_port_port_security_create(context, data, result)

    def process_update_port(self, context, data, result):
        if psec.PORTSECURITY in data:
            self._process_port_port_security_update(
                context, data, result)

    def extend_network_dict(self, session, db_data, result):
        self._extend_port_security_dict(result, db_data)

    def extend_port_dict(self, session, db_data, result):
        self._extend_port_security_dict(result, db_data)

    def _determine_port_security(self, context, port):
        """Returns a boolean (port_security_enabled).

        Port_security is the value associated with the port if one is present
        otherwise the value associated with the network is returned.
        """
        # we don't apply security groups for dhcp, router
        if port.get('device_owner') and net.is_port_trusted(port):
            return False

        if validators.is_attr_set(port.get(psec.PORTSECURITY)):
            port_security_enabled = port[psec.PORTSECURITY]
        else:
            port_security_enabled = self._get_network_security_binding(
                context, port['network_id'])

        return port_security_enabled



# Copyright (c) 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from neutron_lib.api.definitions import port_numa_affinity_policy as apidef
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib.db import api as db_api

from neutron.db import db_base_plugin_v2
from neutron.db import port_numa_affinity_policy_db as pnap_db
from neutron.tests.unit.db import test_db_base_plugin_v2


class PortNumaAffinityPolicyExtensionTestPlugin(
        db_base_plugin_v2.NeutronDbPluginV2,
        pnap_db.PortNumaAffinityPolicyDbMixin):
    """Test plugin to mixin the port NUMA affinity policy extension."""

    supported_extension_aliases = [apidef.ALIAS]

    def create_port(self, context, port):
        with db_api.CONTEXT_WRITER.using(context):
            new_port = super(
                PortNumaAffinityPolicyExtensionTestPlugin,
                self).create_port(context, port)
            self._process_create_port(context, port['port'], new_port)
        return new_port

    def update_port(self, context, id, port):
        with db_api.CONTEXT_WRITER.using(context):
            updated_port = super(
                PortNumaAffinityPolicyExtensionTestPlugin,
                self).update_port(context, id, port)
            updated_port[portbindings.VIF_TYPE] = portbindings.VIF_TYPE_UNBOUND
            self._process_update_port(context, port['port'], updated_port)
        return updated_port


@ddt.ddt
class PortNumaAffinityPolicyExtensionTestCase(
         test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
    """Test API extension numa_affinity_policy attributes."""

    def setUp(self, *args):
        plugin = ('neutron.tests.unit.extensions.test_port_numa_affinity_'
                  'policy.PortNumaAffinityPolicyExtensionTestPlugin')
        super(PortNumaAffinityPolicyExtensionTestCase,
              self).setUp(plugin=plugin)

    def _create_and_check_port_nap(self, numa_affinity_policy):
        name = 'numa_affinity_policy'
        keys = [('name', name), ('admin_state_up', True),
                ('status', self.port_create_status),
                ('numa_affinity_policy', numa_affinity_policy)]
        with self.port(name=name,
                       numa_affinity_policy=numa_affinity_policy) as port:
            for k, v in keys:
                self.assertEqual(v, port['port'][k])
        return port

    def _update_and_check_port_nap(self, port, numa_affinity_policy):
        data = {'port': {'numa_affinity_policy': numa_affinity_policy}}
        req = self.new_update_request('ports', data,
                                      port['port']['id'])
        res = self.deserialize(self.fmt, req.get_response(self.api))
        self.assertEqual(numa_affinity_policy,
                         res['port']['numa_affinity_policy'])

    @ddt.data(*constants.PORT_NUMA_POLICIES, None)
    def test_create_and_update_port_numa_affinity_policy(self,
                                                         numa_affinity_policy):
        port = self._create_and_check_port_nap(numa_affinity_policy)
        for new_nap in (*constants.PORT_NUMA_POLICIES, None):
            self._update_and_check_port_nap(port, new_nap)



# Copyright (c) 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import signal
from unittest import mock

from neutron_lib import constants as n_consts
from neutron_lib import fixture as lib_fixtures
from oslo_utils import uuidutils

from neutron.agent.l3 import ha_router
from neutron.agent.l3 import router_info
from neutron.tests import base
from neutron.tests.common import l3_test_common

_uuid = uuidutils.generate_uuid


class TestBasicRouterOperations(base.BaseTestCase):
    def setUp(self):
        super(TestBasicRouterOperations, self).setUp()
        self.device_exists_p = mock.patch(
            'neutron.agent.linux.ip_lib.device_exists')
        self.device_exists = self.device_exists_p.start()

    def _create_router(self, router=None, **kwargs):
        if not router:
            router = mock.MagicMock()
        self.agent_conf = mock.Mock()
        self.router_id = _uuid()
        return ha_router.HaRouter(mock.sentinel.agent,
                                  self.router_id,
                                  router,
                                  self.agent_conf,
                                  mock.sentinel.driver,
                                  **kwargs)

    def test_get_router_cidrs_returns_ha_cidrs(self):
        ri = self._create_router()
        device = mock.MagicMock()
        device.name.return_value = 'eth2'
        addresses = ['15.1.2.2/24', '15.1.2.3/32']
        ri._get_cidrs_from_keepalived = mock.MagicMock(return_value=addresses)
        self.assertEqual(set(addresses), ri.get_router_cidrs(device))

    def test_routes_updated_with_dvr(self):
        ri = self._create_router(router={'distributed': True})
        ri.keepalived_manager = mock.Mock()
        base_routes_updated = mock.patch(
            'neutron.agent.l3.router_info.'
            'RouterInfo.routes_updated').start()
        mock_instance = mock.Mock()
        mock_instance.virtual_routes.gateway_routes = []
        ri._get_keepalived_instance = mock.Mock(
            return_value=mock_instance)
        ri.routes_updated([], [])
        self.assertTrue(base_routes_updated.called)

    def test_routes_updated_with_non_dvr(self):
        ri = self._create_router(router={'distributed': False})
        ri.keepalived_manager = mock.Mock()
        base_routes_updated = mock.patch(
            'neutron.agent.l3.router_info.'
            'RouterInfo.routes_updated').start()
        mock_instance = mock.Mock()
        mock_instance.virtual_routes.gateway_routes = []
        ri._get_keepalived_instance = mock.Mock(return_value=mock_instance)
        ri.routes_updated([], [])
        self.assertFalse(base_routes_updated.called)

    def test__add_default_gw_virtual_route(self):
        ri = self._create_router()
        mock_instance = mock.Mock()
        mock_instance.virtual_routes.gateway_routes = []
        ri._get_keepalived_instance = mock.Mock(return_value=mock_instance)
        subnets = [{'id': _uuid(),
                    'cidr': '20.0.0.0/24',
                    'gateway_ip': None}]
        ex_gw_port = {'fixed_ips': [],
                      'subnets': subnets,
                      'extra_subnets': [],
                      'id': _uuid(),
                      'network_id': _uuid(),
                      'mac_address': 'ca:fe:de:ad:be:ef'}
        # Make sure no exceptional code
        ri._add_default_gw_virtual_route(ex_gw_port, 'qg-abc')
        self.assertEqual(0, len(mock_instance.virtual_routes.gateway_routes))

        subnets.append({'id': _uuid(),
                        'cidr': '30.0.0.0/24',
                        'gateway_ip': '30.0.0.1'})
        ri._add_default_gw_virtual_route(ex_gw_port, 'qg-abc')
        self.assertEqual(1, len(mock_instance.virtual_routes.gateway_routes))

        subnets[1]['gateway_ip'] = None
        ri._add_default_gw_virtual_route(ex_gw_port, 'qg-abc')
        self.assertEqual(0, len(mock_instance.virtual_routes.gateway_routes))

        subnets[1]['gateway_ip'] = '30.0.1.1'
        ri._add_default_gw_virtual_route(ex_gw_port, 'qg-abc')
        self.assertEqual(2, len(mock_instance.virtual_routes.gateway_routes))

    @mock.patch.object(router_info.RouterInfo, 'remove_floating_ip')
    def test_remove_floating_ip(self, super_remove_floating_ip):
        ri = self._create_router(mock.MagicMock())
        mock_instance = mock.Mock()
        ri._get_keepalived_instance = mock.Mock(return_value=mock_instance)
        device = mock.Mock()
        fip_cidr = '15.1.2.3/32'

        ri.remove_floating_ip(device, fip_cidr)
        self.assertTrue(super_remove_floating_ip.called)

    def test_destroy_state_change_monitor_ok(self):
        ri = self._create_router(mock.MagicMock())
        # need a port for destroy_state_change_monitor() to call PM code
        ri.ha_port = {'id': _uuid()}
        with mock.patch.object(ri,
                               '_get_state_change_monitor_process_manager')\
                as m_get_state:
            mock_pm = m_get_state.return_value
            mock_pm.active = False
            ri.destroy_state_change_monitor(mock_pm)

        mock_pm.disable.assert_called_once_with(
            sig=str(int(signal.SIGTERM)))

    def test_destroy_state_change_monitor_force(self):
        ri = self._create_router(mock.MagicMock())
        # need a port for destroy_state_change_monitor() to call PM code
        ri.ha_port = {'id': _uuid()}
        with mock.patch.object(ri,
                               '_get_state_change_monitor_process_manager')\
                as m_get_state:
            mock_pm = m_get_state.return_value
            mock_pm.active = False
            with mock.patch.object(ha_router, 'SIGTERM_TIMEOUT', 0):
                ri.destroy_state_change_monitor(mock_pm)

        calls = ["sig='str(%d)'" % signal.SIGTERM,
                 "sig='str(%d)'" % signal.SIGKILL]
        mock_pm.disable.has_calls(calls)

    def _test_ha_state(self, read_return, expected):
        ri = self._create_router(mock.MagicMock())
        ri.keepalived_manager = mock.Mock()
        ri.keepalived_manager.get_full_config_file_path.return_value = (
            'ha_state')
        self.mock_open = self.useFixture(
            lib_fixtures.OpenFixture('ha_state', read_return)).mock_open
        self.assertEqual(expected, ri.ha_state)

    def test_ha_state_primary(self):
        self._test_ha_state('primary', 'primary')

    def test_ha_state_unknown(self):
        # an empty state file should yield 'unknown'
        self._test_ha_state('', 'unknown')

    def test_ha_state_ioerror(self):
        # an error reading the state file should yield 'unknown'
        ri = self._create_router(mock.MagicMock())
        ri.keepalived_manager = mock.Mock()
        ri.keepalived_manager.get_full_config_file_path.return_value = (
            'ha_state')
        self.mock_open = IOError
        self.assertEqual('unknown', ri.ha_state)

    def test_gateway_ports_equal(self):
        ri = self._create_router(mock.MagicMock())
        ri.driver = mock.MagicMock()
        subnet_id, qos_policy_id = _uuid(), _uuid()
        _, old_gw_port = l3_test_common.prepare_ext_gw_test(
            self, ri, True)
        old_gw_port['qos_policy_id'] = qos_policy_id
        new_gw_port = copy.deepcopy(old_gw_port)
        new_gw_port.update({'binding:host_id': 'node02',
                            'updated_at': '2018-11-02T14:07:00',
                            'revision_number': 101,
                            'qos_policy_id': qos_policy_id})
        self.assertTrue(ri._gateway_ports_equal(old_gw_port, new_gw_port))

        fixed_ip = {'ip_address': '10.10.10.3', 'subnet_id': subnet_id}
        new_gw_port['fixed_ips'].append(fixed_ip)
        self.assertFalse(ri._gateway_ports_equal(old_gw_port, new_gw_port))

        new_gw_port['fixed_ips'].remove(fixed_ip)
        new_gw_port['qos_policy_id'] = _uuid()
        self.assertFalse(ri._gateway_ports_equal(old_gw_port, new_gw_port))

    def test_set_ha_port(self):
        ri = self._create_router()
        self.assertIsNone(ri.ha_port)

        ri.router = {}
        ri.set_ha_port()
        self.assertIsNone(ri.ha_port)

        # HA_INTERFACE_KEY from None to some value
        ri.router = {n_consts.HA_INTERFACE_KEY: {"id": _uuid(),
                                                 "status": "DOWN"}}
        ri.set_ha_port()
        self.assertIsNotNone(ri.ha_port)
        self.assertEqual('DOWN', ri.ha_port["status"])

        # HA port state change
        ri.router = {n_consts.HA_INTERFACE_KEY: {"id": _uuid(),
                                                 "status": "ACTIVE"}}
        ri.set_ha_port()
        self.assertIsNotNone(ri.ha_port)
        self.assertEqual('ACTIVE', ri.ha_port["status"])

        ri.router = {}
        ri.set_ha_port()
        # neutron server return empty HA_INTERFACE_KEY, but
        # agent side router info should remain the original value.
        self.assertIsNotNone(ri.ha_port)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.callbacks import events as local_events
from neutron_lib.callbacks import registry as local_registry
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging

from neutron.api.rpc.callbacks.consumer import registry
from neutron.api.rpc.callbacks import events
from neutron.api.rpc.callbacks import resources
from neutron.services.trunk.drivers.openvswitch.agent import ovsdb_handler
from neutron.services.trunk.drivers.openvswitch.agent import trunk_manager
from neutron.services.trunk.rpc import agent

LOG = logging.getLogger(__name__)

TRUNK_SKELETON = None


@local_registry.has_registry_receivers
class OVSTrunkSkeleton(agent.TrunkSkeleton):
    """It processes Neutron Server events to create the physical resources
    associated to a logical trunk in response to user initiated API events
    (such as trunk subport add/remove). It collaborates with the OVSDBHandler
    to implement the trunk control plane.
    """

    def __init__(self, ovsdb_handler):
        super(OVSTrunkSkeleton, self).__init__()
        self.ovsdb_handler = ovsdb_handler
        registry.unsubscribe(self.handle_trunks, resources.TRUNK)

    def handle_trunks(self, context, resource_type, trunk, event_type):
        """This method is not required by the OVS Agent driver.

        Trunk notifications are handled via local OVSDB events.
        """
        raise NotImplementedError()

    def handle_subports(self, context, resource_type, subports, event_type):
        # Subports are always created with the same trunk_id and there is
        # always at least one item in subports list
        trunk_id = subports[0].trunk_id

        if self.ovsdb_handler.manages_this_trunk(trunk_id):
            if event_type not in (events.CREATED, events.DELETED):
                LOG.error("Unknown or unimplemented event %s", event_type)
                return

            ctx = self.ovsdb_handler.context
            try:
                LOG.debug("Event %s for subports: %s", event_type, subports)
                if event_type == events.CREATED:
                    status = self.ovsdb_handler.wire_subports_for_trunk(
                            ctx, trunk_id, subports)
                elif event_type == events.DELETED:
                    subport_ids = [subport.port_id for subport in subports]
                    status = self.ovsdb_handler.unwire_subports_for_trunk(
                        trunk_id, subport_ids)
                self.ovsdb_handler.report_trunk_status(ctx, trunk_id, status)
            except oslo_messaging.MessagingException as e:
                LOG.error(
                    "Error on event %(event)s for subports "
                    "%(subports)s: %(err)s",
                    {'event': event_type, 'subports': subports, 'err': e})

    @local_registry.receives(resources.TRUNK, [local_events.BEFORE_CREATE])
    def check_trunk_dependencies(
            self, resource, event, trigger, payload=None):
        # The OVS trunk driver does not work with iptables firewall and QoS.
        # We should validate the environment configuration and signal that
        # something might be wrong.
        # NOTE(armax): this check could be made quite sophisticated in that
        # we could check for incompatibilities and abort the creation request
        # only if the trunk is indeed associated with ports that have security
        # groups and QoS rules, though this would be a lot more work.
        if "iptables_hybrid" in cfg.CONF.SECURITYGROUP.firewall_driver:
            LOG.warning(
                "Firewall driver iptables_hybrid is not compatible with "
                "trunk ports. Trunk %(trunk_id)s may be insecure.",
                {'trunk_id': payload.resource_id})


def init_handler(resource, event, trigger, payload=None):
    """Handler for agent init event."""
    # Set up agent-side RPC for receiving trunk events; we may want to
    # make this setup conditional based on server-side capabilities.
    global TRUNK_SKELETON

    manager = trunk_manager.TrunkManager(trigger.int_br)
    handler = ovsdb_handler.OVSDBHandler(manager)
    TRUNK_SKELETON = OVSTrunkSkeleton(handler)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path
import time

from neutron_lib import constants
import webob
import webob.dec
import webob.exc

from neutron.agent.linux import utils
from neutron.tests.common import machine_fixtures
from neutron.tests.common import net_helpers
from neutron.tests.functional.agent.l3 import framework
from neutron.tests.functional.agent.linux import helpers

METADATA_REQUEST_TIMEOUT = 60
METADATA_REQUEST_SLEEP = 5


class MetadataFakeProxyHandler(object):

    def __init__(self, status):
        self.status = status

    @webob.dec.wsgify()
    def __call__(self, req):
        return webob.Response(status=self.status)


class MetadataL3AgentTestCase(framework.L3AgentTestFramework):

    SOCKET_MODE = 0o644

    def _create_metadata_fake_server(self, status):
        server = utils.UnixDomainWSGIServer('metadata-fake-server')
        self.addCleanup(server.stop)

        # NOTE(cbrandily): TempDir fixture creates a folder with 0o700
        # permissions but metadata_proxy_socket folder must be readable by all
        # users
        self.useFixture(
            helpers.RecursivePermDirFixture(
                os.path.dirname(self.agent.conf.metadata_proxy_socket), 0o555))
        server.start(MetadataFakeProxyHandler(status),
                     self.agent.conf.metadata_proxy_socket,
                     workers=0, backlog=4096, mode=self.SOCKET_MODE)

    def _query_metadata_proxy(self, machine):
        url = 'http://%(host)s:%(port)s' % {'host': constants.METADATA_V4_IP,
                                            'port': constants.METADATA_PORT}
        cmd = 'curl', '--max-time', METADATA_REQUEST_TIMEOUT, '-D-', url
        i = 0
        CONNECTION_REFUSED_TIMEOUT = METADATA_REQUEST_TIMEOUT // 2
        while i <= CONNECTION_REFUSED_TIMEOUT:
            try:
                raw_headers = machine.execute(cmd)
                break
            except RuntimeError as e:
                if 'Connection refused' in str(e):
                    time.sleep(METADATA_REQUEST_SLEEP)
                    i += METADATA_REQUEST_SLEEP
                else:
                    self.fail('metadata proxy unreachable '
                              'on %s before timeout' % url)

        if i > CONNECTION_REFUSED_TIMEOUT:
            self.fail('Timed out waiting metadata proxy to become available')
        return raw_headers.splitlines()[0]

    def test_access_to_metadata_proxy(self):
        """Test access to the l3-agent metadata proxy.

        The test creates:
         * A l3-agent metadata service:
           * A router (which creates a metadata proxy in the router namespace),
           * A fake metadata server
         * A "client" namespace (simulating a vm) with a port on router
           internal subnet.

        The test queries from the "client" namespace the metadata proxy on
        http://169.254.169.254 and asserts that the metadata proxy added
        the X-Forwarded-For and X-Neutron-Router-Id headers to the request
        and forwarded the http request to the fake metadata server and the
        response to the "client" namespace.
        """
        router_info = self.generate_router_info(enable_ha=False)
        router = self.manage_router(self.agent, router_info)
        self._create_metadata_fake_server(webob.exc.HTTPOk.code)

        # Create and configure client namespace
        router_ip_cidr = self._port_first_ip_cidr(router.internal_ports[0])
        br_int = framework.get_ovs_bridge(
            self.agent.conf.OVS.integration_bridge)

        machine = self.useFixture(
            machine_fixtures.FakeMachine(
                br_int,
                net_helpers.increment_ip_cidr(router_ip_cidr),
                router_ip_cidr.partition('/')[0]))

        # Query metadata proxy
        firstline = self._query_metadata_proxy(machine)

        # Check status code
        self.assertIn(str(webob.exc.HTTPOk.code), firstline.split())


class UnprivilegedUserMetadataL3AgentTestCase(MetadataL3AgentTestCase):
    """Test metadata proxy with least privileged user.

    The least privileged user has uid=65534 and is commonly named 'nobody' but
    not always, that's why we use its uid.
    """

    SOCKET_MODE = 0o664

    def setUp(self):
        super(UnprivilegedUserMetadataL3AgentTestCase, self).setUp()
        self.agent.conf.set_override('metadata_proxy_user', '65534')


class UnprivilegedUserGroupMetadataL3AgentTestCase(MetadataL3AgentTestCase):
    """Test metadata proxy with least privileged user/group.

    The least privileged user has uid=65534 and is commonly named 'nobody' but
    not always, that's why we use its uid.
    Its group has gid=65534 and is commonly named 'nobody' or 'nogroup', that's
    why we use its gid.
    """

    SOCKET_MODE = 0o666

    def setUp(self):
        super(UnprivilegedUserGroupMetadataL3AgentTestCase, self).setUp()
        self.agent.conf.set_override('metadata_proxy_user', '65534')
        self.agent.conf.set_override('metadata_proxy_group', '65534')



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ipaddress
import subprocess  # nosec
import sys

import click
import openstack

from neutron._i18n import _


class cached_property:
    __slots__ = ('_fn', '__doc__')

    def __init__(self, fn):
        self._fn = fn
        self.__doc__ = fn.__doc__

    def __get__(self, obj, objtype=None):
        # class of wrapped method must have writable __dict__
        attr = self._fn(obj)
        setattr(obj, self._fn.__name__, attr)
        return attr


class OvnTrace:
    def __init__(self, eth_src_obj, ip_src_obj, eth_dst_obj, ip_dst_obj,
                 extra_flow, extra_args):
        self.inport = eth_src_obj.inport
        self.eth_src = eth_src_obj.mac
        self.ip_src = ip_src_obj.ip
        self.eth_dst = eth_dst_obj.mac
        self.ip_dst = ip_dst_obj.ip
        self.extra_flow = extra_flow
        self.extra_args = extra_args

    @property
    def microflow(self):
        ip_version = ipaddress.ip_address(self.ip_src).version
        generated_flow = (
            'inport == "{inport}" && '
            'eth.src == {eth_src} && '
            'ip{ip_src_v}.src == {ip_src} && '
            'eth.dst == {eth_dst} && '
            'ip{ip_dst_v}.dst == {ip_dst} && '
            'ip.ttl == 64'.format(
                inport=self.inport, eth_src=self.eth_src, ip_src_v=ip_version,
                ip_src=self.ip_src, eth_dst=self.eth_dst, ip_dst_v=ip_version,
                ip_dst=self.ip_dst))
        return ' && '.join(f for f in (generated_flow, self.extra_flow) if f)

    @property
    def args(self):
        return ('ovn-trace', *self.extra_args, self.microflow)

    def run(self):
        return subprocess.run(self.args, check=True)  # nosec

    def __str__(self):
        return " ".join(self.args[:-1] + ("'%s'" % self.args[-1],))


class Interface:
    def __init__(self, query, direction, ctx):
        # Only store the state for running the queries, we don't want to
        # make network connections until after argument parsing is done
        self.query = query
        self.direction = direction
        self.ctx = ctx

    @property
    def network_param(self):
        return (self.ctx.params.get('%s_net' % self.direction) or
                self.ctx.params['net'])

    @property
    def cloud(self):
        return self.ctx.cloud


class ServerInterface(Interface):

    def get_iface(self, iface_type):
        return next(iter(i for i in self.instance.addresses[self.network]
                         if i.get('OS-EXT-IPS:type') == iface_type))

    @cached_property
    def instance(self):
        matching_vms = self.cloud.search_servers(self.query)
        if len(matching_vms) < 1:
            raise Exception(_('Server not found'))
        if len(matching_vms) > 1:
            raise Exception(_('Multiple VMs match %s' % self.query))
        return matching_vms[0]

    @cached_property
    def network(self):
        if self.network_param:
            return self.network_param
        if len(self.instance.addresses) != 1:
            raise Exception(_("Could not determine server network"))
        return next(iter(key for key in self.instance.addresses))

    @cached_property
    def inport(self):
        return self.cloud.search_ports(
            filters={'device_id': self.instance.id})[0].id

    @cached_property
    def ip(self):
        iface = self.get_iface('fixed')
        return iface['addr']

    @cached_property
    def mac(self):
        iface = self.get_iface('fixed')
        return iface['OS-EXT-IPS-MAC:mac_addr']

    @cached_property
    def floating_ip(self):
        iface = self.get_iface('floating')
        return iface['addr']


class RouterInterface(Interface):

    @cached_property
    def network(self):
        return self.cloud.search_networks(self.network_param)[0]

    @cached_property
    def router(self):
        return self.cloud.search_routers(self.query)[0]

    @cached_property
    def interface(self):
        return next(iter(self.cloud.search_ports(
            filters={'device_id': self.router.id,
                     'network_id': self.network.id})))

    @cached_property
    def inport(self):
        return self.interface.id

    @cached_property
    def ip(self):
        return next(iter(ip['ip_address'] for ip in self.interface.fixed_ips))

    @cached_property
    def mac(self):
        return self.interface.mac_address


class SwitchPort(Interface):
    cache = {}


class IP(Interface):
    @property
    def ip(self):
        return self.query


class MAC(Interface):
    @property
    def mac(self):
        return self.query


class RequiredUnless(click.Option):
    def __init__(self, *args, **kwargs):
        try:
            self.unless = kwargs.pop('unless')
            kwargs['required'] = True
        except KeyError:
            raise TypeError(_("Missing required argument: 'unless'"))
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.unless in opts:
            self.required = False
        return super().handle_parse_result(ctx, opts, args)


class ObjEqValueParam(click.ParamType):
    types = {}
    name = "object=value"
    default_type = ""

    def __init__(self, direction):
        if direction not in ('from', 'to'):
            raise ValueError(_("Direction must be 'from' or 'to'"))
        self.direction = direction
        super().__init__()

    def get_metavar(self, param):
        return "[{}]=value".format("|".join(self.types.keys()))

    def convert(self, value, param, ctx):
        if isinstance(value, self.__class__):
            return value
        try:
            obj, value = value.split('=')
        except ValueError:
            obj = self.default_type
        try:
            return self.types[obj](value, self.direction, ctx)
        except KeyError:
            self.fail(f"Unknown object type {obj!r}", param, ctx)


class ToFromParam(ObjEqValueParam):
    types = {'server': ServerInterface, 'router': RouterInterface}
    default_type = 'server'


class MacParam(ObjEqValueParam):
    types = dict(mac=MAC, **ToFromParam.types)
    default_type = "mac"


class IpParam(ObjEqValueParam):
    types = dict(ip=IP, **ToFromParam.types)
    default_type = "ip"


def set_cloud(ctx, param, value):
    ctx.cloud = openstack.connect(cloud=value)
    return ctx.cloud


FromOpt = ToFromParam('from')
ToOpt = ToFromParam('to')
IpSrcOpt = IpParam('from')
EthSrcOpt = MacParam('from')
IpDstOpt = IpParam('to')
EthDstOpt = MacParam('to')


@click.command()
@click.option('--cloud', '-c', default='devstack', callback=set_cloud,
              help='Cloud from clouds.yaml to connect to')
@click.option('--net', '-n', help="Network to limit interfaces lookups to")
@click.option('--from-net', help="Network to limit src interface lookups to")
@click.option('--to-net', help="Network to limit dst interface lookups to")
@click.option('--from', '-f', 'from_', type=FromOpt,
              help='Fill eth-src/ip-src from the same object, e.g. server=vm1')
@click.option('--eth-src', type=EthSrcOpt, cls=RequiredUnless, unless='from_',
              help='Object from which to fill eth.src')
@click.option('--ip-src', type=IpSrcOpt, cls=RequiredUnless, unless='from_',
              help='Object from which to fill ip.src')
@click.option('--to', '-t', type=ToOpt,
              help='Fill eth-dst/ip-dst from the same object, e.g. server=vm2')
@click.option('--eth-dst', '--via', '-v', type=EthDstOpt, cls=RequiredUnless,
              unless='to', help='Object from which to fill eth.dst')
@click.option('--ip-dst', type=IpDstOpt, cls=RequiredUnless, unless='to',
              help='Object from which to fill ip.dst')
@click.option('--microflow', '-m', default='',
              help='Additional microflow text to append to the one generated')
@click.option('--verbose', '-v', is_flag=True, help='Enables verbose mode')
@click.option('--dry-run', is_flag=True,
              help="Print ovn-trace output, but don't run it")
@click.argument('ovntrace_args', nargs=-1, type=click.UNPROCESSED)
def main(cloud, net, from_net, to_net, from_, eth_src, ip_src, to, eth_dst,
         ip_dst, microflow, verbose, dry_run, ovntrace_args):
    ovn_trace = OvnTrace(eth_src or from_, ip_src or from_,
                         eth_dst or to, ip_dst or to, microflow, ovntrace_args)
    if not dry_run:
        if verbose:
            sys.stderr.write("%s\n" % ovn_trace)
        ovn_trace.run()
    else:
        print(ovn_trace)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.objects import common_types
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from neutron_lib.services.qos import constants as qos_constants
from oslo_versionedobjects import fields as obj_fields

from neutron.objects import base


class RuleTypeField(obj_fields.BaseEnumField):

    def __init__(self, **kwargs):
        self.AUTO_TYPE = obj_fields.Enum(
            valid_values=qos_constants.VALID_RULE_TYPES)
        super(RuleTypeField, self).__init__(**kwargs)


@base.NeutronObjectRegistry.register
class QosRuleType(base.NeutronObject):
    # Version 1.0: Initial version
    # Version 1.1: Added QosDscpMarkingRule
    # Version 1.2: Added QosMinimumBandwidthRule
    # Version 1.3: Added drivers field
    # Version 1.4: Added QosPacketRateLimitRule
    # Version 1.5: Added QosMinimumPacketRateRule
    VERSION = '1.5'

    fields = {
        'type': RuleTypeField(),
        'drivers': obj_fields.ListOfObjectsField(
            'QosRuleTypeDriver', nullable=True)
    }

    synthetic_fields = ['drivers']

    # we don't receive context because we don't need db access at all
    @classmethod
    def get_object(cls, rule_type_name, **kwargs):
        plugin = directory.get_plugin(alias=constants.QOS)
        drivers = plugin.supported_rule_type_details(rule_type_name)
        drivers_obj = [QosRuleTypeDriver(
            name=driver['name'],
            supported_parameters=driver['supported_parameters'])
            for driver in drivers]

        return cls(type=rule_type_name, drivers=drivers_obj)

    # we don't receive context because we don't need db access at all
    @classmethod
    def get_objects(cls, validate_filters=True, **kwargs):
        if validate_filters:
            cls.validate_filters(**kwargs)

        rule_types = (
            directory.get_plugin(alias=constants.QOS).supported_rule_types)

        # TODO(ihrachys): apply filters to returned result
        return [cls(type=type_) for type_ in rule_types]

    # we don't receive context because we don't need db access at all
    @classmethod
    def get_values(cls, field, **kwargs):
        return [getattr(obj, field) for obj in cls.get_objects(**kwargs)]


@base.NeutronObjectRegistry.register
class QosRuleTypeDriver(base.NeutronObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'name': obj_fields.StringField(),
        'supported_parameters': common_types.ListOfDictOfMiscValuesField()
    }

    def to_dict(self):
        return {
            'name': self.name,
            'supported_parameters': self.supported_parameters}

    @classmethod
    def get_objects(cls, context, **kwargs):
        raise NotImplementedError()



# Copyright (c) 2012 OpenStack Foundation.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib.api import extensions


class Portbindings(extensions.APIExtensionDescriptor):
    """Extension class supporting port bindings.

    This class is used by neutron's extension framework to make
    metadata about the port bindings available to external applications.

    With admin rights one will be able to update and read the values.
    """

    api_definition = portbindings
